
#include "pcopy-opts.h"

static void
short_read(file_seg_t * fseg, ssize_t rdlen, ssize_t rdct)
{
    static char const fmt[] =
        "th %2d read %d bytes not %d at 0x%08lX\n";
    char msgbf[sizeof(fmt) + 64];
    sprintf(msgbf, fmt,
            fseg->idx, (int)rdct, (int)rdlen, fseg->start);
    fserr(PCOPY_EXIT_FS_ERR_IN, msgbf, fname);
    /* NOTREACHED */
}

static void *
do_copy(void * arg)
{
    file_seg_t * fseg = arg;
    size_t readlim = fseg->end - fseg->start;
    if (readlim > 16 * 1024)
        readlim = 16 * 1024;

    FILE * cur_fp = fopen(fseg->fname, "r");

    if (cur_fp == NULL)
        fserr(PCOPY_EXIT_FS_ERR_IN, "fopen", fseg->fname);

    pthread_mutex_lock(&tty_mutex);
    move(1 + 2 * fseg->idx, 0);
    printw("th %2d reading 0x%08lX thru 0x%08lX",
           fseg->idx, fseg->start, fseg->end);
    refresh();
    pthread_mutex_unlock(&tty_mutex);

    {
        int fdin  = fileno(cur_fp);
        int fdout = open(fseg->seg_name, O_RDWR | O_CREAT, 0600);

        if (fdout < 0)
            fserr(PCOPY_EXIT_FS_ERR_OUT, "open", fseg->seg_name);

        if (fseg->start > 0) {
            if (lseek(fdin,  fseg->start, SEEK_SET) != fseg->start)
                fserr(PCOPY_EXIT_FS_ERR_IN, "seek", fseg->fname);

            if (lseek(fdout, fseg->start, SEEK_SET) != fseg->start)
                fserr(PCOPY_EXIT_FS_ERR_OUT, "seek", fseg->seg_name);
        }

        while (fseg->start < fseg->end) {
            ssize_t rdlen = fseg->end - fseg->start;
            if (rdlen > readlim)
                rdlen = readlim;
            ssize_t rdct = read(fdin, fseg->buffer, rdlen);
            if (rdct <= 0)
                short_read(fseg, rdlen, rdct);

            rdlen = write(fdout, fseg->buffer, rdct);
            if (rdlen != rdct)
                fserr(PCOPY_EXIT_FS_ERR_OUT, "write", fseg->seg_name);

            fseg->start += rdlen;

            pthread_mutex_lock(&tty_mutex);
            move(2 + 2 * fseg->idx, 0);
            printw("     0x%08lX read", fseg->start);
            refresh();
            pthread_mutex_unlock(&tty_mutex);
        }
    }

    pthread_mutex_lock(&tty_mutex);
    move(2 + 2 * fseg->idx, 0);
    clrtoeol();
    move(1 + 2 * fseg->idx, 0);
    clrtoeol();
    printw("th %2d DONE", fseg->idx);
    refresh();
    pthread_mutex_unlock(&tty_mutex);
    free(arg);
    pthread_exit(0);
}

static inline char const *
basenm(char const * f)
{
    char const * res = strrchr(f, '/');
    return res ? (res + 1) : f;
}

file_seg_t *
new_file_seg(char const * fnm, size_t start, size_t seg_len)
{
    static char const too_many[] =
        "The %s destination is a regular file, yet there are multiple inputs";
    static char const bad_dest[] =
        "The %s destination is neither a file nor a directory";

    static bool been_here = false;

    char const * bfile = basenm(fnm);
    file_seg_t * res;
    size_t len = sizeof(*res) + seg_len;

    enum {
        NMIS_UNK,
        NMIS_F_ONLY,
        NMIS_D_ONLY,
        NMIS_F_N_D
    } nm_type = NMIS_UNK;

    if (! HAVE_OPT(DESTINATION)) {
        len += strlen(bfile);
        nm_type = NMIS_F_ONLY;

    } else {
        struct stat sb;
        if (stat(OPT_ARG(DESTINATION), &sb) == 0) {
            if (S_ISDIR(sb.st_mode)) {
                len += strlen(OPT_ARG(DESTINATION)) + strlen(bfile) + 1;
                nm_type = NMIS_F_N_D;

            } else if (S_ISREG(sb.st_mode)) {
                if (been_here)
                    die(PCOPY_EXIT_FAILURE, too_many, OPT_ARG(DESTINATION));
                been_here = true;
                unlink(OPT_ARG(DESTINATION));
                len += strlen(OPT_ARG(DESTINATION));
                nm_type = NMIS_D_ONLY;

            } else {
                die(PCOPY_EXIT_FS_ERR_OUT, bad_dest, OPT_ARG(DESTINATION));
                /* NOTREACHED */
            }
        }
    }

    res = malloc(len);
    if (res == NULL) {
        errno = ENOMEM;
        fserr(PCOPY_EXIT_NO_MEM, "alloc copy space", fnm);
    }

    memset(res, NUL, len);

    switch (nm_type) {
    case NMIS_F_ONLY:
        len = strlen(bfile);
        memcpy(res->seg_name, bfile, ++len);
        break;

    case NMIS_D_ONLY:
        len = strlen(OPT_ARG(DESTINATION));
        memcpy(res->seg_name, OPT_ARG(DESTINATION), ++len);
        break;

    case NMIS_F_N_D:
        len = sprintf(res->seg_name, "%s/%s", OPT_ARG(DESTINATION), bfile) + 1;
        
    default:
        break;
    }

    {
        uintptr_t addr = (uintptr_t)res->seg_name + len + 1 + sizeof(char *);
        addr &= ~(sizeof(char *) - 1);
        res->buffer = (void *)addr;
    }

    res->start  = start;
    res->end    = start + seg_len;
    res->fname  = fnm;
    return res;
}

void
pcopy(char const * fn)
{
    size_t base_size;
    size_t seg_size;
    size_t modulus;
    int    ix = 0;

    fname = fn;

    {
        struct stat sb;
        if (stat(fname, &sb) < 0)
            fserr(PCOPY_EXIT_FAILURE, "cannot re-stat", fname);
        if (! S_ISREG(sb.st_mode)) {
            errno = EINVAL;
            fserr(PCOPY_EXIT_FS_ERR_IN, "regular file check", fname);
        }
        base_size = sb.st_size;
        modulus   = sb.st_blksize;
    }

    {
        size_t sz = base_size / OPT_VALUE_THREAD_CT;

        /*
         * Make sure that thread-ct * size >= base_size
         */
        if ((sz * OPT_VALUE_THREAD_CT) != base_size)
            sz += modulus;

        /*
         * Now make sure sz is an even multiple of modulus
         */
        seg_size = modulus * (sz / modulus);
    }

    memset(pth_list, 0, OPT_VALUE_THREAD_CT * sizeof(*pth_list));

    for (modulus = 0; modulus < base_size; ix++) {
        file_seg_t * fseg = new_file_seg(fname, modulus, seg_size);

        modulus    += seg_size;
        fseg->end   = modulus;
        if (fseg->end > base_size)
            fseg->end = base_size;
        fseg->idx   = ix;
        pthread_create(pth_list + ix, NULL, do_copy, (void *)fseg);
    }

    for (int j = 0; j < ix; j++)
        pthread_join(pth_list[j], NULL);
}
